import gtk, gtk.glade
import singleton
from dialog import GLADEFILE

class PreferencesWin(singleton.SingletonDlg):
    def __init__(self, fg, bg, palette, vte):
        super(PreferencesWin, self).__init__(self)
        self.colors = palette
        self.fg = fg
        self.bg = bg
        self.vte = vte
        self.xml = gtk.glade.XML(GLADEFILE, 'prefdialog')
        self.setup_colorpage(palette)
        self.win = self.xml.get_widget('prefdialog')
        self.win.hide()
        self.xml.signal_autoconnect(self)

    def get_foreground_string(self):
        return self._color_to_string(self.fg)

    def get_background_string(self):
        return self._color_to_string(self.bg)

    def get_palette_string(self):
        return [self._color_to_string(col) for col in self.colors]

    def _color_to_string(self, c):
        s = '#'
        for val in [c.red, c.green, c.blue]:
            s += '%04x' % val #strip leading '0x'
        return s

    def on_okbutton_clicked(self, button):
        self.win.hide()
        
    def on_prefnotebook_switch_page(self, notebook, page, page_num):
        handlers = [self.switch_to_colorpage]
        handlers[page_num](notebook, page, page_num)

    def switch_to_colorpage(self, notebook, page, page_num):
        pass

    def setup_colorpage(self, colors):
        table = self.xml.get_widget('colortable')
        for i, c in enumerate(colors):
            l = gtk.Label(i)
            e = gtk.EventBox()
            e.modify_bg(gtk.STATE_NORMAL, c)
            e.add(l)
            e.connect('button-release-event', self.on_color_activate, i)
            table.attach(e, i/8, i/8+1, i%8+1, i%8+2)
            
    def on_color_activate(self, eventbox, signal_id, i):
        dlg = gtk.ColorSelectionDialog('Color %s'%i)
        dlg.colorsel.set_current_color(self.colors[i])
        dlg.connect('response', self.on_colordlg_response, eventbox, i)
        dlg.present()

    def on_colordlg_response(self, dialog, response_id, eventbox, i):
        dialog.hide()
        if response_id == gtk.RESPONSE_OK:
            color = dialog.colorsel.get_current_color()
            eventbox.modify_bg(gtk.STATE_NORMAL, color)
            self.colors[i] = color
            self.vte.set_colors(self.fg, self.bg, self.colors)
