#!/usr/bin/python
import os
import signal
import gtk, gtk.glade, pango, vte, gobject

from singleton import SingletonDlg
from dialog import GLADEFILE
import config
from keybinding import Keybinding
import colors
import about
import preferences

HOSTFIELDS=['name', 'address', 'port']
HOSTTYPES =[str,    str,       int   ]
CONTEXT_DESC = 'Server messages'
class PyTermixApp(gobject.GObject):
    def __init__(self):
        '''Set up the app itself.'''

        super(PyTermixApp, self).__init__()

        #Create the window, but hide it so we can change the size and things
        #without the user seeing.
        self.xml = gtk.glade.XML(GLADEFILE, 'window1')
        self.xml.signal_autoconnect(self)
        menubar = self.xml.get_widget('menubar1')

        menus = menubar.get_children()
        for menu in menus:
            print menu.get_submenu().get_accel_group()

        win = self.xml.get_widget('window1')
        win.hide()

        vbox = self.xml.get_widget('eventbox1')
        c = vbox.get_colormap().alloc_color("dark gray")
        vbox.modify_bg(gtk.STATE_NORMAL, c)

        accelgroup = gtk.AccelGroup()
        def f(*args):
            print args
        accelgroup.connect_group(ord('s'), gtk.gdk.MOD1_MASK, gtk.ACCEL_VISIBLE,
                                 f)
                                 
        #win.add_accelerator('clicked', accelgroup, ord('s'),
        #                    gtk.gdk.MOD1_MASK, gtk.ACCEL_VISIBLE)
        win.add_accel_group(accelgroup)

        #Resize the window -- resize is necessary, because by the time
        #libglade is done, the window has already been sized.
        win.resize(*config.size)
        #win.set_default_size(*config.size)


        #Fit the VTETerminal widget in the scrolledwindow left for us in Glade.
        w = self.xml.get_widget('vtetarget')
        v = vte.Terminal()
        #w.pack_start(v, fill=False)
        w.add(v)
        v.show()

        #Sync the scrolledwindow's adjustment to the VTETerminal's adjustment
        #so that you can scroll the widget with the scrollbar.
        #w.set_vadjustment(v.get_adjustment())

        #ANSI drawing characters show up in the cp437 encoding, and the widget
        #looks best with the FreeMono font.
        v.set_encoding('cp437')
        v.set_font(pango.FontDescription('FreeMono 16'))
        v.grab_focus()
        v.set_size(80, 24)
        print v.get_size_request(), tuple(v.allocation)

        
        #Set up colors for VTETerminal widget.
        #TODO: Gets colors from colors module, rather than config
        #Also, could have a "preferences" window to select colors..
        map = v.get_colormap()
        for color in [colors.foreground, colors.background]+ colors.palette:
            map.alloc_color(color)
        
        v.set_colors(colors.foreground, colors.background, colors.palette)

        p = preferences.PreferencesWin(colors.foreground, colors.background,
                                       colors.palette, v)

        v.connect('child-exited', self.child_exited)
        v.connect('key-press-event', self.handle_key)

        #Keep track of various parts of the GUI, because we change them
        print 'getting prefs'
        self.prefs = p
        self.connect_to = self.xml.get_widget('connect_to1')
        self.terminal = v
        self.statusbar = self.xml.get_widget('statusbar1')
        self.statuslabel = self.xml.get_widget('statuslabel')

        cid = self.statusbar.get_context_id(CONTEXT_DESC)
        self.statusbar.push(cid, "Not connected")

        self.serverlist = gtk.ListStore(*HOSTTYPES)
        self.load_servers(config.servers)
        self.serverlist.connect('row-changed', self.generate_menu)
        self.serverlist.connect('row-deleted', self.generate_menu)
        self.current_server = [None]
        self.generate_menu()

        self.keybindings = config.keybindings

        win.show()

    def load_servers(self, serverlist):
        for s in config.servers:
            #If we just get (name, address), append -1 as a port,
            #meaning "Default"
            if len(s) == 2: s = s+(-1,)
            iter = self.serverlist.append(s)
            

    def generate_menu(self, *args):
        '''Generate the submenu for the connect_to menu item from the
        liststore of servers.'''
        m = gtk.Menu()
        self.connect_to.set_submenu(m)
        for n, s in enumerate(self.serverlist):
            i = gtk.MenuItem(label="_%d: %s" % (n, s[0]))
            i.connect('activate', self.connect_menuitem, s)
            m.append(i)
            i.show()

        #Add a last menu item for "Make a new connection"
        i = gtk.MenuItem(label='_New Connection...')
        i.connect('activate', self.new_connect)
        i.show()
        m.append(i)
        
    def on_copy1_activate(self, menuitem):
        self.terminal.copy_clipboard()

    def on_paste1_activate(self, menuitem):
        self.terminal.paste_clipboard()

    def on_preferences1_activate(self, menuitem):
        preferences.PreferencesWin(None, self.terminal).show()
    
    def on_connectionslist_activate(self, menuitem):
        ConnectionList(self, self.serverlist).show()

    def on_keybindings_activate(self, menuitem):
        #Need to find out which keybindings to open the Keybinding window on.
        server = self.current_server[0] # name of the server we're connected to
        if server == None:
            server = 0
        KeybindingWin(self, self.serverlist, self.keybindings, server)\
                            .show()

    def connect_menuitem(self, menuitem, server):
        self.connect_to_server(server)
        
    def connect_to_server(self, server):
        '''Connect to a server. Called either by the ConnectionList window, or
        by the "Connect To..." menu items.

        Forks off a telnet command as the child of the terminal.'''
        print "connecting to", server
        print server[1], server[2]
        self.current_server = server

        #Build the args list -- vte.Terminal.fork_command takes arguments like
        #execv, so start it with the program's name.
        args = ['telnet', server[1]]
        if server[2] != -1 and server[2] != 23:
            args.append(server[2])
        id = self.terminal.fork_command('/usr/bin/telnet',
                                        args)
        cid = self.statusbar.get_context_id(CONTEXT_DESC)
        self.statusbar.pop(cid)
        self.statusbar.push(cid, "Connected to %s" % (server[1],))
        self.statuslabel.set_text(server[0])

        self.telnetid = id
        self.terminal.grab_focus()

    def child_exited(self, terminal):
        cid = self.statusbar.get_context_id(CONTEXT_DESC)
        self.statusbar.pop(cid)
        self.statusbar.push(cid, "Disconnected from %s." % self.current_server[1])
        self.statuslabel.set_text("Disconnected")
        #self.last_server = self.current_server #Is this necessary?
        
    def new_connect(self, menuitem):
        EditServerWin(self, None)
        
    def on_reconnect1_activate(self, menuitem):
        self.connect_to_server(self.current_server)

    def on_disconnect1_activate(self, menuitem):
        if hasattr(self, 'telnetid'):
            os.kill(self.telnetid, signal.SIGKILL)
        self.child_exited(None)
        
    def on_about1_activate(self, menuitem):
        about.AboutDlg().show()
    

    def handle_key(self, terminal, event):
        '''Handle a keystroke.

        If it matches a macro, return True, to stop further processing of event.'''
        key = self.check_macro(event)
        if key == False:
            return False
        else:
            key = key.replace('\n', '\r')
            terminal.feed_child(key)
            print key
            return True

    def check_macro(self, event):
        '''Check to see if event matches a macro.

        If so, return it, to stop further processing of event.
        '''
        assert event.type == gtk.gdk.KEY_PRESS

        k = Keybinding(event)
        print "Here's the keybindings so far"
        print self.keybindings

        #First, check global keybindings
        if k in self.keybindings[0]: #bindings[0] == global bindings
            return self.keybindings[0][k]


        #Then, check server-specific bindings
        #Current server name = self.current_server[0]
        #Bindings for that name is self.keybindings[self.current_server[0]]
        #Binding for key k would be self.keybindings[self.current_server[0]][k]
        if self.current_server[0] not in self.keybindings:
            #No binding found. return False to continue processing
            return False

        if k in self.keybindings[self.current_server[0]]:
            return self.keybindings[self.current_server[0]][k]

        else:
            #We didn't find a binding -- the terminal should do its thing
            return False 

    def quit(self, *args):
        "Quits the program, saving current state to .pytermix"
        output = []
        output.append("size = %s\n" % (self.xml.get_widget('window1').get_size(),))

        #Builds up the list of servers, so as to join them with string.join
        #Each entry is built from an entry in the list store.
        l = []
        for x in self.serverlist:
            #Compute this entry. Join the fields together with ', ' and add
            #braces.
            l.append('[' + ', '.join([`i` for i in x]) + ']')
        output.append('servers = [' + ',\n\t'.join(l) + ']\n')
        output.append('keybindings = %s\n' % (self.keybindings))
        output.append('foreground = "%s"\n' % self.prefs.get_foreground_string())
        output.append('background = "%s"\n' % self.prefs.get_background_string())
        output.append('colors = %s\n' % (self.prefs.get_palette_string()))
        gtk.main_quit()
        #After here, open the user's file. DON'T RAISE ANYTHING, or user data
        # will be lost!
        out = file(os.path.expanduser('~/.pytermix'), 'w')
        out.writelines(output)

    def edit_ok(self, iter, fields):
        '''Callback for when a server was edited or added.'''
        list = self.serverlist
        if iter is None:
            iter = list.append()

        for n, v in enumerate(fields):
            list.set(iter, n, HOSTTYPES[n](v))

gobject.signal_new('hotkey', PyTermixApp, gobject.SIGNAL_ACTION, None,
                   (gtk.gdk.Event,))


class ConnectionList(SingletonDlg):
    def __init__(self, parent, list):
        self.xml = gtk.glade.XML(GLADEFILE, 'conndialog')
        self.xml.signal_autoconnect(self)
        self.parent = parent
        self.win = self.xml.get_widget('conndialog')
        self.setup_treeview(list)

    def setup_treeview(self, list):
        tv = self.xml.get_widget('treeview1')
        tv.set_model(list)
        for n, x in enumerate(HOSTFIELDS):
            x = x.capitalize()
            tv.append_column(gtk.TreeViewColumn(x, gtk.CellRendererText(),
                                                text=n))

    def on_connokbutton_clicked(self, button):
        self.win.hide()

    #def on_conncancelbutton_clicked(self, button):
    #    self.win.hide()
    #Button deleted for now


    def on_treeview1_row_activated(self, treeview, path, view_column):
        self.win.hide()
        self.parent.connect_to_server(treeview.get_model()[path])


    def on_newbutton_clicked(self, button):
        EditServerWin(self, None)

    def selected(self):
        #returns pair: list, iter
        return self.xml.get_widget('treeview1').get_selection().get_selected()
    
    #Beware: local variable "list" shadows builtin "list" (i.e. []) below
    def on_editbutton_clicked(self, button):
        (list, iter) = self.selected()
        EditServerWin(self, iter, *list.get(iter, *range(len(HOSTFIELDS))))

    def on_delbutton_clicked(self, button):
        (list, iter) = self.selected()
        list.remove(iter)

    def on_downbutton_clicked(self, button):
        self.move_selected(0)

    def on_upbutton_clicked(self, button):
        self.move_selected(1)

    def move_selected(self, up = 0):
        'Moves the currently selected treerow up or down, depending on arg.'
        (listm, iter) = self.selected()
        path = list(listm.get_path(iter)) #tuple -> list so we can modify
        #Modify the last element of the path only
        if up == 1:
            path[-1] -= 1
        else:
            path[-1] += 1
        iter2 = listm.get_iter(tuple(path))
        listm.swap(iter, iter2)

    def edit_ok(self, iter, fields):
        #Just pass this up to PytermixApp.edit_ok
        self.parent.edit_ok(iter, fields)

class EditServerWin(object):
    def __init__(self, parent, iter, name='', address='', port = 23):
        '''Creates a window for editing a server's name, address and port.

        parent is the window which will be alerted when the edit is completed.

        iter is the iter of the liststore that should be changed if the edit
        is successful. None means add a new entry.'''
        self.xml = gtk.glade.XML(GLADEFILE, 'editdialog')
        self.xml.signal_autoconnect(self)
        self.win = self.xml.get_widget('editdialog')
        self.parent = parent
        self.iter = iter
        #Get attributes straight from the locals, i.e. function arguments
        #At the same time, set up the correspondingly-named textentries
        for x in HOSTFIELDS:
            print 'set', x
            setattr(self, x, locals()[x])
            self.xml.get_widget(x+'entry').set_text(str(locals()[x]))

    def on_editokbutton_clicked(self, button):
        fields = []
        for x in HOSTFIELDS:
            fields.append(self.xml.get_widget(x+'entry').get_text())
        self.parent.edit_ok(self.iter, fields)
        self.win.hide()

    def on_editcancelbutton_clicked(self, button):
        self.win.hide()
        del self

KEYTYPES = [str, str, object]
KEYNAMES = ['key', 'action']
GLOBAL = 'Global bindings'
class KeybindingWin(SingletonDlg):
    '''Class representing a GUI window for editing keybindings.

    Has only a "Close" button, rather than an Ok/Cancel button.
    Accordingly, any changes made to a keybinding take effect immediately.
    Is this a bug, or feature?
    Whichever it is, ConnectionList is the same way.
    '''
    def __init__(self, parent, liststore, bindings, current):
        print 'wank'
        self.parent = parent
        self.serverlist = gtk.ListStore(str) #For the combobox
        self.update_servers(liststore)
        liststore.connect('row-changed', self.update_servers)
        liststore.connect('row-deleted', self.update_servers)
        
        self.liststore = gtk.ListStore(*KEYTYPES) #For the treeview
        self.bindings = bindings
        self.current_server = current
        self.xml = gtk.glade.XML(GLADEFILE, 'keydialog')
        self.win = self.xml.get_widget('keydialog')
        self.setup_entry(self.serverlist)
        self.setup_treeview()
        self.xml.signal_autoconnect(self)

    def update_servers(self, liststore, *args):
        #Copy liststore, and add a "global" entry
        self.serverlist.clear()
        self.serverlist.append([GLOBAL])
        for x in liststore:
            print x
            self.serverlist.append([x[0]])
            
    def show(self):
        super(KeybindingWin, self).show()
        self.fill_bindings()

    def setup_entry(self, liststore):
        entry = self.xml.get_widget('keyserverlist')
        entry.set_model(liststore)
        cell = gtk.CellRendererText()
        entry.pack_start(cell)
        entry.add_attribute(cell, 'text', 0)
        liststore = entry.get_model()
        entry.set_active_iter(liststore.get_iter("0"))
        self.fill_bindings(0)
        #    entry.append_text(s[0])
        #    if self.current == s:
        #        entry.set_active(n)

    def setup_treeview(self):
        tv = self.xml.get_widget('keytreeview')
        tv.set_model(self.liststore)
        for n, x in enumerate(KEYNAMES):
            x = x.capitalize()
            col = gtk.TreeViewColumn(x, gtk.CellRendererText(),
                                     text=n)
            col.set_property('min-width', 100)
            tv.append_column(col)

    def fill_bindings(self, server = None):
        print "Fill_bindings(", server, ")"
        if server is None:
            server = self.current_server
        if server == GLOBAL:
            server = 0
        print server
        if server not in self.bindings:
            self.bindings[server] = {}
        liststore = self.liststore
        liststore.clear()
        for k,v in self.bindings[server].items():
            liststore.append([k, v, k])

    def on_closebutton1_clicked(self, button):
        self.win.hide()

    def selected(self):
        #FIXME -- similar to ConnectionList.selected
        # Short enough that it doesn't matter?
        #returns pair: list, iter
        return self.xml.get_widget('keytreeview').get_selection().get_selected()

    def combo_selected_iter(self):
        combo = self.xml.get_widget('keyserverlist')
        return combo.get_active_iter()

    
    def on_keynewbutton_clicked(self, button):
        EditKeyWin(self, self.serverlist, self.combo_selected_iter())

    def on_keyeditbutton_clicked(self, button):
        l, i = self.selected()
        EditKeyWin(self, self.serverlist, self.combo_selected_iter(), *l.get(i, 2, 1))
        # field 2 is the keybinding itself

    def on_keydelbutton_clicked(self, button):
        store, iter = self.selected()
        binding = store.get_value(iter, 2)
        del self.bindings[self.current_server][binding]
        store.remove(iter)
        

    def add_binding(self, server, key, text):
        if server == GLOBAL:
            server = 0
        if server not in self.bindings:
            self.bindings[server] = {}
        self.bindings[server][key] = text
        if server == self.current_server:
            self.fill_bindings()

    def on_keyserverlist_changed(self, combo):
        self.current_server = combo.get_active_text() 
        print "Serverlist changed: New = ", self.current_server
        if self.current_server == GLOBAL:
            self.current_server = 0
        self.fill_bindings()

class EditKeyWin(object):
    '''A window allowing a user to edit a single key binding.

    Arguments to the constructor: The parent widget, the list of servers
    to choose from, the iterator specifying which server is default,
    and a key and binding if one exists.
    '''
    def __init__(self, parent, serverlist, serveriter,
                 key = None, binding = None):
        #FIXME: These next lines appear in every dialog __init__ -- fix
        self.xml = gtk.glade.XML(GLADEFILE, 'keyeditdialog')
        self.win = self.xml.get_widget('keyeditdialog')
        self.xml.signal_autoconnect(self)

        combo = self.xml.get_widget('servercombo')
        combo.set_model(serverlist)
        combo.set_active_iter(serveriter)
        cell = gtk.CellRendererText()
        combo.pack_start(cell)
        combo.add_attribute(cell, 'text', 0)

        self.parent = parent
        self.key = key
        self.xml.get_widget('keyentry').set_text(str(self.key))
        self.binding = binding
        if binding:
            self.xml.get_widget('actiontext').get_buffer().set_text(binding)

    def on_keyentry_event(self, entry, event):
        entry.set_text(str(self.key))
        print 'got event'
        print event.type
        if event.type == gtk.gdk.KEY_PRESS:
            print gtk.gdk.keyval_name(event.keyval)
            if gtk.gdk.keyval_name(event.keyval) == "Tab":
                #No keybindings on "Tab"
                return False
            self.key = Keybinding(event)
            return True

    def selected_server(self):
        combo = self.xml.get_widget('servercombo')
        return combo.get_model(), combo.get_active_iter()

    def on_keyeditok_clicked(self, button):
        key = self.key
        tb = self.xml.get_widget('actiontext').get_buffer()
        text = tb.get_text(tb.get_start_iter(), tb.get_end_iter())
        store, iter = self.selected_server()
        server = store.get_value(iter, 0)
        self.parent.add_binding(server, key, text)
        self.close()

    def on_keyeditcancel_clicked(self, button):
        self.close()
        
    def close(self):
        self.win.hide()
        del self.win
        del self

p = PyTermixApp()
gtk.main()
